#!/usr/bin/env python
# -*- coding:utf-8 -*-

import sys
import argparse
from tccli.log import init
import six
from difflib import get_close_matches
from gettext import gettext
from tccli.error_msg import USAGE

log = init("tccli.argparser")

class CustomAction(argparse.Action):

    def __init__(self, option_strings, dest, command_map, **kwargs):
        self.command_map = command_map
        super(CustomAction, self).__init__(option_strings, dest, choices=self.choices, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)

    @property
    def choices(self):
        return list(self.command_map.keys())

    @choices.setter
    def choices(self, val):
        pass


class BaseArgParser(argparse.ArgumentParser):
    Formatter = argparse.RawTextHelpFormatter

    ChoicesPerLine = 2

    def _check_value(self, action, value):
        if action.choices is not None and value not in action.choices:
            msg = ['Invalid choice, valid choices are:\n']
            args = (' '.join(sys.argv[1:])).split(value)
            help_msg = "you can run `tccli %shelp` to see valid choices" % args[0]
            log.error(USAGE + msg[0] + help_msg)
            for i in range(len(action.choices))[::self.ChoicesPerLine]:
                current = []
                for choice in action.choices[i:i + self.ChoicesPerLine]:
                    current.append('%-40s' % choice)
                msg.append(' | '.join(current))
            possible = get_close_matches(value, action.choices, cutoff=0.8)
            if possible:
                extra = ['\n\nInvalid choice: %r, maybe you meant:\n' % value]
                for word in possible:
                    extra.append('  * %s' % word)
                msg.extend(extra)
            msg.append("Invalid choice: The specified command or option is not found, " \
                       "try lastest version by running `pip install --upgrade tccli`.")
            raise argparse.ArgumentError(action, '\n'.join(msg))

    def parse_known_args(self, args=None, namespace=None):
        parsed, remaining = super(BaseArgParser, self).parse_known_args(args, namespace)
        terminal_encoding = getattr(sys.stdin, 'encoding', 'utf-8')
        if terminal_encoding is None:
            terminal_encoding = 'utf-8'
        for arg, value in vars(parsed).items():
            if isinstance(value, six.binary_type):
                setattr(parsed, arg, value.decode(terminal_encoding))
            elif isinstance(value, list):
                encoded = []
                for v in value:
                    if isinstance(v, six.binary_type):
                        encoded.append(v.decode(terminal_encoding))
                    else:
                        encoded.append(v)
                setattr(parsed, arg, encoded)
        return parsed, remaining

    def error(self, message):
        self.print_usage(sys.stderr)
        args = {'prog': self.prog, 'message': message}
        self.exit(252, gettext('%(prog)s: error: %(message)s\n') % args)


class CLIArgParser(BaseArgParser):

    def __init__(self, command_map, version, description, argument_map, prog=None):
        super(CLIArgParser, self).__init__(
            formatter_class=self.Formatter,
            add_help=False,
            conflict_handler='resolve',
            description=description,
            usage=USAGE,
            prog=prog)
        self._build(command_map, version, argument_map)

    def _create_choice_help(self, choices):
        help_str = ''
        for choice in sorted(choices):
            help_str += '* %s\n' % choice
        return help_str

    def _build(self, command_map, version, argument_map):
        for argument_name in argument_map:
            argument = argument_map[argument_name]
            argument.add_to_parser(self)
        self.add_argument('--version',
                          action="version",
                          version=version,
                          help='Display the version of this tool')
        self.add_argument('command', action=CustomAction, command_map=command_map)


class ActionArgParser(BaseArgParser):

    def __init__(self, actions_map):
        super(ActionArgParser, self).__init__(
            formatter_class=self.Formatter,
            add_help=False,
            conflict_handler='resolve',
            usage=USAGE)
        self._build(actions_map)

    def _build(self, actions_map):
        self.add_argument('operation', action=CustomAction,
                          command_map=actions_map)


class ArgMapArgParser(BaseArgParser):

    def __init__(self, argument_map, command_map=None):
        super(ArgMapArgParser, self).__init__(
            formatter_class=self.Formatter,
            add_help=False,
            usage=USAGE,
            conflict_handler='resolve')
        self._build(argument_map, command_map)

    def _build(self, argument_map, command_map):
        for arg_name in argument_map:
            argument = argument_map[arg_name]
            argument.add_to_parser(self)
        if command_map:
            self.add_argument('subcommand', action=CustomAction, command_map=command_map, nargs='?')

    def parse_known_args(self, args=None, namespace=None):
        if len(args) > 0 and args[0] == 'help':
            namespace = argparse.Namespace()
            namespace.help = 'help'
            return namespace, args[1:]
        else:
            return super(ArgMapArgParser, self).parse_known_args(
                args, namespace)



# -*- coding: utf-8 -*-

HELP = (
    "To tccli help text, you can run:\n"
    "\n"
    "  tccli help\n"
    "  tccli configure help\n"
    "  tccli service[cvm] help\n"
    "  tccli service[cvm] action[RunInstances] help\n"
)
USAGE = (
    "tccli [options] <command> <subcommand> [<subcommand> ...] [parameters]\n"
    "%s" % HELP
)



# -*- coding: utf-8 -*-

import os
import six
import copy
import json
from tccli.utils import Utils
from tccli import __version__
from tccli.services import SERVICE_VERSIONS
from collections import OrderedDict

BASE_TYPE = ["int64", "uint64", "string", "float", "bool", "date", "datetime", "datetime_iso", "binary"]
CLI_BASE_TYPE = ["Integer", "String", "Float", "Timestamp", "Boolean", "Binary"]

PARAM_TYPE_MAP = {
    'int64': 'Integer',
    'uint64': 'Integer',
    'int': 'Integer',
    'string': 'String',
    'float': 'Float',
    'bool': 'Boolean',
    'date': "Timestamp",
    'datetime': "Timestamp",
    'datetime_iso': "Timestamp",
    'binary': 'Binary',
    'object': 'Object',
    'list': 'Array'
}

HELPER_MAP = {
    "--profile": "specify a profile name",
    "--secretId": "specify a SecretId",
    "--secretKey": "specify a SecretKey",
    "--token": "temporary certificate token",
    "--role-arn": "specify a RoleArn",
    "--role-session-name": "specify a RoleSessionName",
    "--use-cvm-role": "use CVM Role to obtain the secret id and secret key",
    "--region": "identify the region to which the instance you want to work with belongs.",
    "--endpoint": "specify an access point domain name",
    "--detail": "see detailed help information",
    "--filter": "specify a filter field",
    "--timeout": "specify a request timeout",
    "--generate-cli-skeleton": "Prints a JSON skeleton to standard output without sending "
                               "an API request. If provided with no value or the value "
                               "input, prints a sample input JSON that can be used as an "
                               "argument for --cli-input-json. If provided with the value "
                               "output, it validates the command inputs and returns a "
                               "sample output JSON for that command.",
    "--cli-input-json": "Reads arguments from the JSON string provided. The JSON string "
                        "follows the format provided by --generate-cli-skeleton. ",
    "--cli-unfold-argument": "complex type parameters are expanded with dots",
    "--https-proxy": "specify a https proxy",
    "--warning": "Open the warning log",
    "--waiter": "Set param `expr`, `to`, `timeout` and `interval` to get the polling result."
                "`expr` is the inquiry expresion, `to` is the ending status"
                ".`timeout` and `interval` are optional params.",
    "--language": "specify an output language, valid choices: [zh-CN, en-US], default value: zh-CN",
}

class Loader(object):
    def get_services_path(self):
        base_path = os.path.dirname(os.path.abspath(__file__))
        services_path = os.path.join(base_path, 'services')
        return services_path

    def get_cli_version(self):
        return __version__

    def get_description(self):
        return "tccli (Tencent Cloud Command Line Interface) is a tool to manage your Tencent Cloud services."

    def get_configure(self):
        return "Before using tccli, you should use the command(tccli configure) to configure your profile " \
               "as the default For more information, please enter tccli configure help"

    def get_usage(self):
        return "tccli [options] <service> [options] <action> [options] [options and parameters]"

    def get_options(self):
        return {
            "help": "show the tccli help info",
            "--version": "show the version of tccli"
        }

    def get_cli_option(self):
        return {
            "filter": {
                "help": HELPER_MAP['--filter'],
            },
            "output": {
                "choices": [
                    "json",
                    "text",
                    "table"
                ],
                "metavar": "output_format"
            },
            "secretId": {
                "help": HELPER_MAP['--secretId'],
            },
            "secretKey": {
                "help": HELPER_MAP['--secretKey'],
            },
            "token": {
                "help": HELPER_MAP['--token'],
            },
            "role-arn": {
                "help": HELPER_MAP['--role-arn'],
            },
            "role-session-name": {
                "help": HELPER_MAP['--role-session-name'],
            },
            "use-cvm-role": {
                "help": HELPER_MAP['--use-cvm-role'],
                'action': 'store_true'
            },
            "version": {
                "help": "specify a DescribeRegions api version",
                "metavar": "version_name"
            },
            "detail": {
                "help": HELPER_MAP['--detail'],
                'action': 'store_true'
            },
            "profile": {
                "help": HELPER_MAP['--profile'],
                "metavar": "profile_name"
            },
            "region": {
                "help": HELPER_MAP['--region'],
                "metavar": "region_name"
            },
            "endpoint": {
                "help": HELPER_MAP['--endpoint'],
                "metavar": "endpoint_url"
            },
            "timeout": {
                "type": "int",
                "help": HELPER_MAP['--timeout'],
            },
            "generate-cli-skeleton": {
                'help': HELPER_MAP['--generate-cli-skeleton'],
                'nargs': '?',
                'const': 'input',
                'choices': ['input'],
            },
            'cli-input-json': {
                'help': HELPER_MAP['--cli-input-json'],
            },
            'cli-unfold-argument': {
                'help': HELPER_MAP['--cli-unfold-argument'],
                'action': 'store_true'
            },
            'https-proxy': {
                'help': HELPER_MAP['--https-proxy'],
            },
            'warning': {
                'help': HELPER_MAP['--warning'],
                'action': 'store_true'
            },
            'waiter': {
                'help': HELPER_MAP['--waiter'],
            },
            "language": {
                'help': HELPER_MAP['--language'],
                "choices": [
                    "zh-CN",
                    "en-US"
                ],
            },
        }

    def _version_transform(self, version):
        return version[1:5] + "-" + version[5:7] + "-" + version[7:9]

    def get_available_services(self):
        return SERVICE_VERSIONS

    def get_service_default_version(self, service):
        return self.get_available_services()[service][0]

    def get_service_model(self, service, version):
        services_path = self.get_services_path()
        version = "v" + version.replace('-', '')
        apis_path = os.path.join(services_path, service, version, "api.json")
        if not os.path.exists(apis_path):
            raise Exception("Not find service:%s version:%s model" % (service, version))

        if six.PY2:
            with open(apis_path, 'r') as f:
                return json.load(f)
        else:
            with open(apis_path, 'r', encoding='utf-8') as f:
                return json.load(f)

    def get_service_description(self, service, version):
        service_model = self.get_service_model(service, version)
        description = service_model["metadata"].get("api_brief", '')
        return description

    def get_service_usage(self, service):
        return "tccli %s <action> [--param...]" % service

    def get_service_options(self, service):
        return {
            "help": "show the tccli %s help info" % service
        }

    def get_action_model(self, service, version, action):
        service_model = self.get_service_model(service, version)
        action_model = service_model["actions"][action]
        return action_model

    def get_actions_info(self, service, version):
        service_model = self.get_service_model(service, version)
        actions_info = OrderedDict()
        for action in sorted(service_model["actions"]):
            actions_info[action] = {}
            actions_info[action]["name"] = service_model["actions"][action]["name"]
            actions_info[action]["document"] = service_model["actions"][action]["document"]
        return actions_info

    def get_service_all_version_actions(self, service):
        services_path = self.get_services_path()
        module_path = os.path.join(services_path, service)
        if not os.path.exists(module_path):
            raise Exception("Not find service:%s " % service)
        version_actions = {}
        for version in os.listdir(module_path):
            if version.startswith("v") and os.path.isdir(os.path.join(module_path, version)):
                version = self._version_transform(version)
                actions = self.get_actions_info(service, version).keys()
                version_actions[version] = actions
        return version_actions

    def get_service_all_action_param(self, service, model=None):
        version_actions = self.get_service_all_version_actions(service)
        version_action_params = {}
        for version in version_actions:
            version_action_params[version] = {}
            for action in version_actions[version]:
                if model == "cli-unfold-argument":
                    params = self.get_unfold_param_info(service, version, action).keys()
                else:
                    params = self.get_param_info(service, version, action).keys()
                version_action_params[version][action] = params
        action_params = {}
        for version in version_action_params:
            for action in version_action_params[version]:
                if action in action_params:
                    action_params[action] = list(set(action_params[action] + version_action_params[version][action]))
                else:
                    action_params[action] = version_action_params[version][action]
        return action_params

    def get_action_description(self, service, version, action):
        return self.get_actions_info(service, version)[action]['document']

    def get_action_usage(self, service, action):
        return "tccli %s %s [--param...]" % (service, action)

    def get_action_options(self, service, action):
        HELPER_MAP["help"] = "show the tccli %s %s help info" % (service, action)
        HELPER_MAP["--version"] = "specify a %s api version" % action
        return HELPER_MAP

    def _filling_param_info(self, param_info, para, param_type, member):
        param_info[para["name"]] = {}
        param_info[para["name"]]["document"] = para["document"]
        if "required" in para:
            param_info[para["name"]]["required"] = "Required" if para["required"] else "Optional"
        if "value_allowed_null" in para:
            param_info[para["name"]]["value_allowed_null"] = \
                "AllowedNull" if para["value_allowed_null"] else "NotAllowedNull"
        param_info[para["name"]]["type_name"] = PARAM_TYPE_MAP.get(para["member"], para["member"])
        param_info[para["name"]]["type"] = PARAM_TYPE_MAP.get(param_type, param_type)
        param_info[para["name"]]["members"] = member
        return param_info

    def _get_param_info(self, param_model, object_model):
        param_info = {}
        for para in param_model:
            if para["type"] == "list":
                if para["member"] not in BASE_TYPE:
                    self._filling_param_info(
                        param_info, para, "list",
                        [self._get_param_info(object_model[para["member"]]["members"], object_model)])
                else:
                    self._filling_param_info(
                        param_info, para, "list", [para["member"]])
            else:
                if para["member"] not in BASE_TYPE:
                    param_info = self._filling_param_info(
                        param_info, para, para["member"],
                        self._get_param_info(object_model[para["member"]]["members"], object_model))
                else:
                    self._filling_param_info(param_info, para, para["member"], para["member"])
        return param_info

    def get_param_info(self, service, version, action):
        service_model = self.get_service_model(service, version)
        param_model = service_model["objects"]
        return self._get_param_info(param_model[action + "Request"]["members"], param_model)

    def get_output_param_info(self, service, version, action):
        service_model = self.get_service_model(service, version)
        param_model = service_model["objects"]
        return self._get_param_info(param_model[action + "Response"]["members"], param_model)

    def _generate_param_skeleton(self, param_model, name):
        param_skeleton = {}
        for para in param_model:
            if para["type"] == "list":
                if para["member"] not in BASE_TYPE:
                    param_skeleton[para["name"]] = \
                        [self._generate_param_skeleton(name[para["member"]]["members"], name)]
                else:
                    param_skeleton[para["name"]] = [PARAM_TYPE_MAP[para["member"]]]
            else:
                if para["member"] not in BASE_TYPE:
                    param_skeleton[para["name"]] = \
                        self._generate_param_skeleton(name[para["member"]]["members"], name)
                else:
                    param_skeleton[para["name"]] = PARAM_TYPE_MAP[para["member"]]
        return param_skeleton

    def generate_param_skeleton(self, service, version, action):
        service_model = self.get_service_model(service, version)
        param_model = service_model["objects"]
        return self._generate_param_skeleton(param_model[action + "Request"]["members"], param_model)

    def get_unfold_param_info(self, service, version, action, profile="default", param_array=False):
        service_model = self.get_service_model(service, version)
        object_model = service_model["objects"]
        all_param_list = []
        for para in object_model[action + "Request"]["members"]:
            param_list = []
            self._get_unfold_param_info(object_model, all_param_list, param_list, para)

        if param_array:
            all_param_list = self._add_array_item(all_param_list, profile)

        return self._filling_unfold_param_info(all_param_list, service, version, action)

    def _add_array_item(self, param_list, profile):
        is_conf_exist, conf_path = Utils.file_existed(os.path.join(os.path.expanduser("~"), ".tccli"),
                                                      profile + ".configure")
        if is_conf_exist:
            array_count = Utils.load_json_msg(conf_path).get("arrayCount", 10)
        else:
            array_count = 10
        all_param_list = param_list
        for para in param_list:
            for idx, item in enumerate(para):
                if item == '0':
                    for i in range(1, int(array_count)):
                        tmp = copy.deepcopy(para)
                        tmp[idx] = str(i)
                        all_param_list.append(tmp)
        return all_param_list

    def _recur_get_unfold_param_info(self, param_model, object_model, return_param_list, param_list):
        for para in param_model:
            self._get_unfold_param_info(object_model, return_param_list, param_list, para)
        if param_list.pop().isdigit():
            param_list.pop()

    def _get_unfold_param_info(self, object_model, return_param_list, param_list, para):
        param_list.append(para["name"])
        if para["type"] == "list" and para["member"] not in BASE_TYPE:
            param_list.append('0')
        if para["member"] not in BASE_TYPE:
            self._recur_get_unfold_param_info(object_model[para["member"]]["members"],
                                              object_model, return_param_list, param_list)
        else:
            tmp = copy.deepcopy(param_list)
            return_param_list.append(tmp)

            if param_list.pop().isdigit():
                param_list.pop()

    def _filling_unfold_param_info(self, param_list, service, version, action):
        unfold_param = {}
        param_info = self.get_param_info(service, version, action)
        for param in param_list:
            unfold_param[".".join(param)] = {}

            tmp_param = [item for item in param if not item.isdigit()]
            res = param_info[tmp_param[0]]

            param_type = res["type"]
            type_name = res["type_name"]
            required = res["required"]
            document = res["document"]

            for idx, item in enumerate(tmp_param[1:]):
                if res["type"] == "Array":
                    res = res["members"][0][item]
                else:
                    res = res["members"][item]

                if required == "Required" and res["required"] == "Optional":
                    required = "Optional"

                if idx == len(tmp_param) - 2:
                    param_type = res["type"]
                    type_name = res["type_name"] if res["type"] == "Array" \
                        else res["type_name"]
                    document = res["document"]
                    break

            if len([item for item in param if item.isdigit() and int(item) > 0]) > 0:
                required = "Optional"

            unfold_param[".".join(param)]["type"] = param_type
            unfold_param[".".join(param)]["type_name"] = type_name
            unfold_param[".".join(param)]["required"] = required
            unfold_param[".".join(param)]["document"] = document
        return unfold_param

    def get_action_example_model(self, service, version, action):
        services_path = self.get_services_path()
        version = "v" + version.replace('-', '')
        example_path = os.path.join(services_path, service, version, "examples.json")
        if not os.path.exists(example_path):
            raise Exception("Not find service:%s version:%s model" % (service, version))

        with open(example_path, 'r') as f:
            examples = json.load(f)
        return examples['actions'][action]

    def generate_cli_example(self, service, version, action):
        examples = self.get_action_example_model(service, version, action)
        for example in examples:
            example["input"] = self.translate_cli_example(service, action, example)
            try:
                example["output"] = json.dumps(json.loads(example["output"], object_pairs_hook=OrderedDict),
                                               indent=4, separators=(',', ': '), ensure_ascii=False)
            except Exception:
                pass
        return examples

    def translate_cli_example(self, module, action, example):
        if example["input"].startswith("https"):
            input_param_list = example["input"].replace(u"&<公共请求参数>", "").split("&")[1:]
            return self.translate_get_cli_param(input_param_list)
        elif example["input"].startswith("POST"):
            input_param = example["input"].split('\n\n')[-1]
            try:
                input_param = json.loads(input_param, object_pairs_hook=OrderedDict)
            except Exception as err:
                raise Exception("service: %s, action: %s, err: %s" % (module, action, err))
            return self.translate_post_cli_param(input_param)

    def translate_get_cli_param(self, input_param_list):
        param_list = [k.strip() for k in input_param_list]
        param_dict = OrderedDict()
        for item in param_list:
            try:
                key = item.split("=")[0]
                value = item.split("=")[1]
            except Exception:
                continue
            if key.split(".")[-1].isdigit():
                key = ".".join(key.split(".")[:-1])
                if key not in param_dict:
                    param_dict[key] = [value]
                else:
                    param_dict[key].append(value)
            else:
                param_dict[key] = value

        cli_param_list = []
        for key, value in param_dict.items():
            cli_param = "--" + key
            if isinstance(value, list):
                for item in value:
                    if " " in str(item):
                        cli_param += " " + "'" + str(item) + "'"
                    else:
                        cli_param += " " + str(item)
            else:
                if " " in str(value):
                    cli_param += " " + "'" + str(value) + "'"
                else:
                    cli_param += " " + str(value)
            cli_param_list.append(cli_param)
        return cli_param_list

    def translate_post_cli_param(self, input_param):
        all_param_list = []
        param_list = []
        self._translate_post_cli_param(input_param, param_list, all_param_list)
        cli_param_list = ["--" + ".".join(item[:-1]) + ' ' + item[-1] for item in all_param_list]
        return cli_param_list

    def _translate_post_cli_param(self, input_param, param_list, all_param_list):
        # basic type
        if not isinstance(input_param, list) and not isinstance(input_param, dict):
            if isinstance(input_param, str) and " " in input_param:
                input_param = "'" + input_param + "'"
            param_list.append(str(input_param))
            tmp = copy.deepcopy(param_list)
            all_param_list.append(tmp)
            param_list.pop()
        # basic type array
        elif isinstance(input_param, list) and len(input_param) > 0 and not isinstance(input_param[0], dict):
            value = " ".join(["'"+str(param)+"'" if " " in str(param) else str(param) for param in input_param])
            param_list.append(value)
            tmp = copy.deepcopy(param_list)
            all_param_list.append(tmp)
            param_list.pop()
        # complex object type array
        elif isinstance(input_param, list) and len(input_param) > 0 and isinstance(input_param[0], dict):
            for idx, param in enumerate(input_param):
                param_list.append(str(idx))
                self._translate_post_cli_param(param, param_list, all_param_list)
                param_list.pop()
        # complex object type
        elif isinstance(input_param, dict):
            for param in input_param:
                param_list.append(param)
                self._translate_post_cli_param(input_param[param], param_list, all_param_list)
                param_list.pop()
        # null array
        else:
            pass




# -*- coding: utf-8 -*-

import json
import os
from tccli.exceptions import ParamError, ParamSyntaxError
from tccli.argument import CustomArgument


class CliInputArgument(CustomArgument):

    def add_to_call_parameters(self, parsed_args, **kwargs):
        call_parameters = {}
        arg_value = self._get_arg_value(parsed_args)
        if arg_value is None:
            return

        loaded_params = self._load_parameters(arg_value)
        if not isinstance(loaded_params, dict):
            raise ParamError(
                self.cli_name,
                "Invalid type: expecting dict, "
                "received %s" % type(loaded_params)
            )
        self._update_call_parameters(call_parameters, loaded_params)
        return call_parameters

    def _get_arg_value(self, parsed_args):
        arg_value = getattr(parsed_args, self.py_name, None)
        if arg_value is None:
            return

        cli_input_args = [
            k for k, v in vars(parsed_args).items()
            if v is not None and k.startswith('cli_input')
        ]
        if len(cli_input_args) != 1:
            raise ParamSyntaxError(
                'Only one --cli-input- parameter may be specified.'
            )
        paramfile_data = self._get_paramfile(arg_value, 'file://')
        if paramfile_data is not None:
            return paramfile_data

        return arg_value

    def _get_paramfile(self, path, prefix):
        data = None
        if path.startswith(prefix):
            file_path = os.path.expandvars(os.path.expanduser(path[len(prefix):]))
            try:
                with open(file_path, 'r') as f:
                    data = f.read()
            except (OSError, IOError) as e:
                raise Exception('Unable to load paramfile %s: %s' % (file_path, e))
        else:
            raise Exception("--cli-input-json parameter must be "
                            "followed by a value beginning with 'file://'\n")
        return data

    def _load_parameters(self, arg_value):
        raise NotImplementedError('_load_parameters')

    def _update_call_parameters(self, call_parameters, loaded_parameters):
        for input_key in loaded_parameters.keys():
            if input_key not in call_parameters:
                call_parameters[input_key] = loaded_parameters[input_key]


class CliInputJSONArgument(CliInputArgument):

    ARG_DATA = {
        'name': 'cli-input-json',
        'group_name': 'cli_input',
        'help_text': (
            'Reads arguments from the JSON string provided. The JSON string '
            'follows the format provided by ``--generate-cli-skeleton``. '
        )
    }

    def __init__(self):
        super(CliInputJSONArgument, self).__init__(**self.ARG_DATA)

    def _load_parameters(self, arg_value):
        try:
            return json.loads(arg_value)
        except ValueError:
            raise ParamError(self.name, "Invalid JSON received.")





import json
from tccli.exceptions import ParamError
from tccli.loaders import CLI_BASE_TYPE


class BaseArgument(object):

    def __init__(self, name):
        self._name = name

    def add_to_arg_map(self, argument_map):
        argument_map[self.name] = self

    def add_to_parser(self, parser):
        pass

    def add_to_params(self, parameters, value):
        pass

    @property
    def name(self):
        return self._name

    @property
    def cli_name(self):
        return '--' + self._name

    @property
    def cli_type_name(self):
        raise NotImplementedError("cli_type_name")

    @property
    def required(self):
        raise NotImplementedError("required")

    @property
    def documentation(self):
        raise NotImplementedError("documentation")

    @property
    def cli_type(self):
        raise NotImplementedError("cli_type")

    @property
    def py_name(self):
        return self._name.replace('-', '_')

    @property
    def choices(self):
        return None

    @property
    def synopsis(self):
        return ''

    @property
    def positional_arg(self):
        return False

    @property
    def nargs(self):
        return None

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def group_name(self):
        return None


class CustomArgument(BaseArgument):

    def __init__(self, name, help_text='', dest=None, default=None,
                 action=None, required=None, choices=None, nargs=None,
                 cli_type_name=None, group_name=None, positional_arg=False,
                 no_paramfile=False, synopsis='',
                 const=None):
        self._name = name
        self._help = help_text
        self._dest = dest
        self._default = default
        self._action = action
        self._required = required
        self._nargs = nargs
        self._const = const
        self._cli_type_name = cli_type_name
        self._group_name = group_name
        self._positional_arg = positional_arg
        self._choices = [] if choices is None else choices
        self._synopsis = synopsis

        self.no_paramfile = no_paramfile

    @property
    def cli_arg_name(self):
        if self._positional_arg:
            return self._name
        else:
            return '--' + self._name

    def add_to_parser(self, parser):
        cli_arg_name = self.cli_arg_name
        kwargs = {}
        if self._dest is not None:
            kwargs['dest'] = self._dest
        if self._action is not None:
            kwargs['action'] = self._action
        if self._default is not None:
            kwargs['default'] = self._default
        if self._choices:
            kwargs['choices'] = self._choices
        if self._required is not None:
            kwargs['required'] = self._required
        if self._nargs is not None:
            kwargs['nargs'] = self._nargs
        if self._const is not None:
            kwargs['const'] = self._const
        parser.add_argument(cli_arg_name, **kwargs)

    @property
    def required(self):
        if self._required is None:
            return False
        return self._required

    @required.setter
    def required(self, value):
        self._required = value

    @property
    def documentation(self):
        return self._help

    @property
    def cli_type(self):
        cli_type = str
        if self._action in ['store_true', 'store_false']:
            cli_type = bool
        return cli_type

    @property
    def choices(self):
        return self._choices

    @property
    def group_name(self):
        return self._group_name

    @property
    def synopsis(self):
        return self._synopsis

    @property
    def positional_arg(self):
        return self._positional_arg

    @property
    def nargs(self):
        return self._nargs


class CLIArgument(BaseArgument):
    TYPE_MAP = {
        'Array': str,
        'String': str,
        'Float': float,
        'Integer': int,
        'Timestamp': str,
        'Boolean': str
    }

    def __init__(self, name, argument_model, action_model,
                 is_required=False):
        self._name = name
        self.argument_model = argument_model
        self._required = is_required
        self._operation_model = action_model
        self._documentation = argument_model["document"]

    @property
    def name(self):
        return self._name

    @property
    def required(self):
        return self._required

    @required.setter
    def required(self, value):
        self._required = value

    @property
    def documentation(self):
        return self._documentation

    @documentation.setter
    def documentation(self, value):
        self._documentation = value

    @property
    def cli_type_name(self):
        return ""

    @property
    def cli_type(self):
        return self.TYPE_MAP.get(self.argument_model["type"], str)

    def add_to_parser(self, parser):
        cli_name = self.cli_name
        parser.add_argument(
            cli_name,
            help=self.documentation,
            type=self.cli_type,
            required=self.required)

    def add_to_params(self, parameters, value):
        if value is None:
            return
        else:
            value = self._unpack_param(value)
            parameters[self._name] = value

    def _unpack_param(self, value):
        if self.argument_model["type"] == "Boolean":
            try:
                if str(value) == 'False':
                    return False
                elif str(value) == 'True':
                    return True
                else:
                    raise ValueError
            except ValueError as e:
                raise ParamError(
                    self.cli_name, "Invalid Boolean: %s\n "
                                   "Boolean received: %s\n "
                                   "Boolean only take values: True/False"
                                   % (e, value))
        elif self.argument_model["type"] in CLI_BASE_TYPE:
            return value
        else:
            try:
                return json.loads(value.encode("utf8"))
            except ValueError as e:
                raise ParamError(
                    self.cli_name, "Invalid JSON: %s\n "
                                   "JSON received: %s" % (e, value))


class ListArgument(CLIArgument):

    @property
    def nargs(self):
        return '*'

    @property
    def cli_type(self):
        return self.TYPE_MAP.get(self.argument_model["type_name"], str)

    def add_to_parser(self, parser):
        cli_name = self.cli_name
        parser.add_argument(cli_name,
                            nargs=self.nargs,
                            type=self.cli_type,
                            required=self.required)


class BooleanArgument(CLIArgument):
    def __init__(self, name, argument_model, action_model,
                 is_required=False, action='store_true', dest=None,
                 group_name=None, default=None):
        super(BooleanArgument, self).__init__(name,
                                              argument_model,
                                              action_model,
                                              is_required)
        self._action = action
        if dest is None:
            self._destination = self.py_name
        else:
            self._destination = dest
        if group_name is None:
            self._group_name = self.name
        else:
            self._group_name = group_name
        self._default = default

    def add_to_params(self, parameters, value):
        if value is not None:
            parameters[self.name] = value

    def add_to_arg_map(self, argument_map):
        argument_map[self.name] = self

    def add_to_parser(self, parser):
        parser.add_argument(self.cli_name,
                            help=self.documentation,
                            action=self._action,
                            default=self._default,
                            dest=self._destination)

    @property
    def group_name(self):
        return self._group_name
